"""
Tunneled SSH client module

Usage:
- {cmd} ssh <deviceId> [ssh-arguments...]
- {cmd} ssh <user>@<deviceId> [ssh-arguments...]


Examples:
- {cmd} ssh root@test
  connect to SSH on your `test` device as user `root`
"""


from ondevice.modules import ModuleClient, ModuleService

import errno
import logging
import socket
import subprocess
import sys

info = 'Connect to your devices\' SSH server'
encrypted = True

# DEPRECATED - use `ondevice ssh` instead
class Client(ModuleClient):
    """ Endpoint stub that simply invokes 'ssh' with the ProxyCommand set to
    'onclient connect ssh:tunnel' """

    def runLocal(self, *args):
        info = self._info
        devId = info.devId
        protocol = info.protocol
        svcName = info.service

        # TODO use the dynamic module name
        proxyCmd = [ sys.argv[0], 'connect', '{0}:tunnel@{1}'.format(protocol, svcName), devId ]

        ssh = subprocess.Popen(['ssh', '-o', 'ProxyCommand={0}'.format(' '.join(proxyCmd)), 'ondevice:{0}'.format(devId)]+list(args), stdin=None, stdout=None, stderr=None)
        return ssh.wait()

    def startRemote(self):
        pass # we don't need a remote connection; Client_tunnel does that for us

class Client_tunnel(ModuleClient):
    def onMessage(self, data):
        logging.debug("onMessage: %s", repr(data))
        stream = self.getConsoleBuffer(sys.stdout)
        stream.write(data)
        stream.flush()

    def onClose(self):
        logging.debug('ondevice ssh: connection closed, exiting')
        sys.exit(0)

    def onEOF(self):
        logging.debug('ondevice ssh: got EOF, closing stdout')
        self.getConsoleBuffer(sys.stdout).close()


    def runLocal(self):
        while True:
            # read1() only invokes the underlying read function only once (and
            # in contrast to read() returns as soon as there's data available,
            # not just when 4096 bytes have actually been read)
            stream = self.getConsoleBuffer(sys.stdin)
            data = stream.read1(4096)
            if data:
                logging.debug("sndData: %s", repr(data))
                self._conn.send(data)
            else:
                logging.info("Local EOF, closing connection")
                self._conn.sendEOF()
                return

class Service(ModuleService):
    def runLocal(self):
        """ Connect to the local SSH server (on port 22) and send all data we get from it over the tunnel """
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # TODO make host+port configurable
        # TODO use a timeout; raise an exception on error
        self._sock.connect(('localhost', 22))
        logging.debug("connected to SSH server: {0} -> {1}".format(self._sock.getsockname(), self._sock.getpeername()))

        while True:
            try:
                data = self._sock.recv(4096)
                if data:
                    logging.debug("sndData: %s", repr(data))
                    self._conn.send(data)
                else:
                    self._conn.sendEOF()
                    return
            except OSError as e:
                if e.errno == errno.EBADF:
                    # socket has been closed (i.e. by onClose())
                    break
                else: raise e

    def stopLocal(self):
        """ stops the loop in runLocal() by closing the SSH socket """
        self._sock.shutdown(socket.SHUT_RDWR)
        self._sock.close()

    def onClose(self):
        ModuleService.onClose(self)
        logging.debug("closing SSH client socket ({0} -> {1})".format(self._sock.getsockname(), self._sock.getpeername()))
        self._sock.shutdown(socket.SHUT_RDWR)
        self._sock.close()

    def onMessage(self, data):
        logging.debug("onMessage: %s", repr(data))
        self._sock.send(data)



import ondevice
from ondevice.core import config, exception, sock

import base64
import json
import logging

try:
    from http.client import HTTPConnection, HTTPSConnection
except ImportError:
    from httplib import HTTPConnection, HTTPSConnection

def apiDELETE(endpoint, params={}, data=None):
    return apiRequest('DELETE', endpoint, params=params, data=data)

def apiGET(endpoint, params={}):
    return apiRequest('GET', endpoint, params=params)

def apiPOST(endpoint, params={}, data=None):
    return apiRequest('POST', endpoint, params=params, data=data)

def apiRequest(method, endpoint, params={}, data=None):
    # TODO implement URL params support
    auth = config.getClientAuth()
    headers = {
        'User-agent': 'ondevice v{0}'.format(ondevice.getVersion())
    }

    if auth != None:
        basicAuth = base64.b64encode("{0}:{1}".format(*auth).encode('ascii')).decode('ascii')
        headers['Authorization'] = 'Basic {0}'.format(basicAuth)

    # parse data:
    if data != None:
        data = json.dumps(data)
        headers.update({
            'Content-type': 'application/json; charset=utf8',
        }) # TODO check if we need to also set the content-length

    if sock.BASE_URL.scheme == 'ws':
        c = HTTPConnection(sock.BASE_URL.hostname, sock.BASE_URL.port)
    elif sock.BASE_URL.scheme == 'wss':
        c = HTTPSConnection(sock.BASE_URL.hostname, sock.BASE_URL.port)
    # TODO do proper URL handling (urllib)
    c.request(method, '/v1.1{0}'.format(endpoint), body=data, headers=headers)
    resp = c.getresponse()
    rc = resp.read()

    if resp.status < 200 or resp.status >= 300:
        # TODO implement HTTP redirect support
        logging.warning("Error message: %s", rc)
        raise exception.TransportError("API server responded with code {0}!".format(resp.status))
    elif resp.getheader('content-type') == None:
        raise exception.TransportError("Response lacks Content-type header!")

    cType=resp.getheader('content-type').lower().split(';')
    if cType[0] != 'application/json':
        raise exception.TransportError("Expected an 'application/json' response (was: '{0}')".format(cType[0]))
    if type(rc) == bytes:
        rc = rc.decode('utf8') # TODO check the actual 'charset' part from cType
    return json.loads(rc)



from ondevice.core import config, exception, service, thread
from ondevice.core.tunnel import Connection, Response, TunnelInfo

import imp
import io
import logging
import os
import pkgutil
import re

class Endpoint:
    def __init__(self, info):
        self.info = info

    def getConsoleBuffer(self, stream):
        """ Compatibilty code for py2/3 (returns stream.buffer in py3 and io.BufferedReader/BufferedWriter in py2) """
        if hasattr(stream, 'buffer'):
            return stream.buffer # py3
        else:
            mode = stream.mode
            if mode in ['r','w']:
                mode = mode+'b'
            elif mode in ['rb','wb']:
                pass
            else:
                raise exception.ImplementationError("Unexpected stream mode: {0}".format(mode))

            rc = io.open(stream.fileno(), mode=mode) # py2
            if hasattr(rc, 'read'):
                rc.read1 = rc.read # this is hacky but it seems to work
            return rc

    def startRemote(self):
        threadName = 'conn_{0}:remote'.format(self.info.connId)
        self._remoteThread = thread.BackgroundThread(self.runRemote, self.stopRemote, name=threadName)
        self._remoteThread.start()

    def startLocal(self):
        threadName = 'conn_{0}:local'.format(self.info.connId)
        args = self._args if hasattr(self, '_args') else ()

        self._localThread = thread.BackgroundThread(self.runLocal, self.stopLocal, name=threadName, args=args)
        self._localThread.start()

    def onMessage(self, data):
        raise exception.ImplementationError("This module doesn't implement the 'onMessage' method!?!")

    def runRemote(self):
        self._conn.run()

    def stopRemote(self):
        """ Closes the tunnel connection """
        self._conn.close()

class ModuleClient(Endpoint):
    def __init__(self, info, args):
        Endpoint.__init__(self, info)
        self._args = args
        self._conn = Connection(info, listener=self)

    def onClose(self):
        raise exception.ImplementationError("onClose not implemented!")

    def onEOF(self):
        raise exception.ImplementationError("onEOF not implemented!")

class ModuleService(Endpoint):
    def __init__(self, info):
        Endpoint.__init__(self, info)
        self._conn = Response(info, listener=self)

    def onClose(self):
        logging.info("Tunnel closed (bytes sent=%d, received=%d)", self.info.bytesSent, self.info.bytesReceived)

def exists(name):
    return name in listModules()

def listModules():
    knownNames = []
    for importer, modName, isPkg in pkgutil.iter_modules(__path__):
        if not isPkg:
            knownNames.append(modName)
            yield modName

    # list user-local modules (stored in .config/ondevice/modules)
    userPath = config._getConfigPath('modules')
    for importer, modName, isPkg in pkgutil.iter_modules([userPath]):
        if not isPkg and modName not in knownNames:
            yield modName

def load(name):
    try:
        ondevice = __import__('ondevice.modules.{0}'.format(name))
        return getattr(ondevice.modules, name)
    except ImportError as e:
        # maybe there's a user-installed module?
        # TODO use importlib if possible
        userPath = config._getConfigPath('modules')
        mod = imp.load_source('usermodules.{0}'.format(name), os.path.join(userPath, '{0}.py'.format(name)))
        return mod

def loadClient(devId, protocolStr, args):
    modName, suffix, svcName = _parseProtocolString(protocolStr)
    if svcName == None:
        svcName = modName
    return loadClient2(devId=devId, modName=modName, suffix=suffix, svcName=svcName, args=args)

def loadClient2(devId, modName, suffix, svcName, args):
    className = 'Client'

    if suffix != None:
        className = '{0}_{1}'.format('Client', suffix)
    mod = load(modName)

    if not hasattr(mod, className):
        raise exception.UsageError("Module '{0}' doesn't have a '{1}' endpoint".format(modName, suffix))
    clazz = getattr(mod, className)
    info=TunnelInfo(devId=devId, protocol=modName, service=svcName, clientUser=config.getClientUser())
    rc = clazz(info, args=args)
    return rc

def getService(req, devKey):
    svc = service.get(req.service)
    if svc['protocol'] != req.protocol:
        raise Exception("Error: protocol mismatch (expected={0}, actual={1})".format(svc['protocol'], req.protocol))
    mod = load(req.protocol)
    info = TunnelInfo(devKey=devKey, tunnelId=req.tunnelId, brokerUrl=req.broker,
            service=req.service, protocol=req.protocol, clientUser=req.clientUser)
    rc = mod.Service(info)
    return rc

def _parseProtocolString(protocolStr):
    """ Takes a protocol string (in the format `moduleName[:suffix][@svcName]`)
    and returns its parts (or None for each missing part) """
    return re.match('([^:@]+):?([^@]+)?@?(.+)?', protocolStr).groups()



from ondevice.core import config, sock, thread, state

import logging
import threading
import time

class TunnelInfo:
    """ Contains information on a tunnel connection.
    Mandatory attributes:
    - service
    - protocol
    - devId/clientUser (depending on whether we're used on the client or device side)
    
    Optional attributes:
    - bytesSent
    - bytesReceived
    - clientUser """

    def __init__(self, **kwargs):
        # make sure we have at least the following arguments:
        for expectedKeys in 'devId/clientUser,service,protocol'.split(','):
            found = False
            for key in expectedKeys.split('/'):
                if key in kwargs:
                    found = True
                    break
                
            if not found:
                raise Exception("Missing TunnelInfo argument: '{0}' (only got: {1})".format(expectedKeys, kwargs.keys()))

        self.bytesSent = 0
        self.bytesReceived = 0
        self.connectedAt = time.time()

        # set the object attributes
        for k, v in kwargs.items():
            setattr(self, k, v)


class TunnelSocket(sock.Socket):
    """ Base class for Connection and Response. """
    def __init__(self, endpoint, info, auth, listener=None, **params):
        if listener != None:
            self.listener = listener
        self._eof = False
        self._lock = threading.Lock()
        self._lock.acquire() # lock initially (will be released once the server confirms the connection)
        self.lastPing = None
        self.info = info
        self.info.connId = state.add('tunnels', 'seq', 1)

        baseUrl = info.brokerUrl if hasattr(info, 'brokerUrl') else None
        sock.Socket.__init__(self, endpoint, auth, baseurl=baseUrl, **params)

        taskName = 'conn_{0}:ping'.format(info.connId)
        self._pingTask = thread.FixedDelayTask(self._ping, 60, name=taskName)

    def _callListener(self, method, *args, **kwargs):
        if self.listener != None:
            if hasattr(self.listener, method):
                getattr(self.listener, method)(*args, **kwargs)
            else:
                raise Exception("missing callback: {0}".format(method))

    def _hasListener(self, method):
        return (self.listener != None
            and hasattr(self.listener, method)
            and callable(getattr(self.listener, method)))

    def _onMessage(self, ws, messageData):
        self.info.bytesReceived += len(messageData)
        colonPos = messageData.find(b':')
        if colonPos < 0:
            raise Exception('Missing message header')

        # 'header' examples:
        # - 'data:': data, channel 0
        # - 'data.3:': data, channel 3
        # - 'meta': metadata
        msgType, messageData = self._takeHeader(messageData)
        logging.debug('{0} << {1} ({2} bytes)'.format(msgType, repr(messageData), len(messageData)))

        if msgType == b'data':
            return self._callListener('onMessage', messageData)
        elif msgType == b'meta':
            msgType, messageData = self._takeHeader(messageData)

            if msgType == b'connected':
                if messageData != None:
                    params = self._parseParams(messageData)
                    assert b'api' in params
                else:
                    params = {b'api': b'v1.0'}

                logging.debug("-- connected --")
                self._apiVersion = params[b'api']
                self._onConnected()
            elif msgType == b'EOF':
                logging.debug("-- got EOF --")
                self.onEOF()
            elif msgType == b'ping':
                logging.debug("-- got ping message --")
                self.lastPing = time.time()

                pongMsg = b'meta:pong'
                if messageData != None:
                    pongMsg += messageData
                self._ws.send(pongMsg, 2) #OPCODE_BINARY
            elif msgType == b'pong':
                self.lastPing = time.time()
            else:
                raise Exception("Unsupported meta message: '{0}'".format(messageData))
        elif msgType == b'error':
            code, messageData = self._takeHeader(messageData)
            code = int(code)
            self.onError(code, messageData.decode('utf8'))
        else:
            raise Exception("Unsupported message type: '{0}'".format(msgType))

    def _onConnected(self):
        self._lock.release()
        self._pingTask.start()

        if self._hasListener('onConnected'):
            self._callListener('onConnected')

    def _onClose(self, ws):
        self._pingTask.stop()
        # TODO we should probably relock self._lock here
        if not self._eof:
            self.onEOF()
        if self._hasListener('onClose'):
            self._callListener('onClose')
        else:
            sock.Socket._onClose(self, ws)

    def _parseParams(self, params):
        """ Parse parameters in the format 'par1=val1,par2=val2' """
        rc = {}
        for param in params.split(b','):
            k,v = param.split(b'=')
            rc[k] = v
        return rc

    def _ping(self):
        """ check that the last ping has been received within the last five minutes.
        These checks will only start after at least one ping/pong was received.
        That way older versions of the ondevice client can coexist with newer ones."""
        if self.lastPing != None:
            now = time.time()
            logging.debug("last ping received: {0}s ago".format(now - self.lastPing))
            if now - self.lastPing > 300:
                self.onError(-1, 'Connection lost')
                self._ws.close()
                raise Exception("Connection timed out :(")

    def _takeHeader(self, msg):
        colonPos = msg.find(b':')
        hdr = None
        if colonPos >= 0:
            hdr = msg[:colonPos]
            msg = msg[colonPos+1:]
        else:
            hdr = msg
            msg = None
        return hdr, msg

    def onEOF(self):
        if self._eof == False:
            self._eof = True
            self._callListener('onEOF')

    def onError(self, code, msg):
        if self._hasListener('onError'):
            self._callListener('onError', code, msg)
        else:
            raise Exception("Error (code={0}): {1}".format(code, msg))

    def send(self, msg):
        if (type(msg) != bytes):
            raise Exception("expected bytes, not {0}".format(type(msg)))

        with self._lock:
            logging.debug('{0} >> {1} ({2} bytes)'.format('data', msg, len(msg)))
            data = b'data:'+msg

            self.info.bytesSent += len(data)
            self._ws.send(data, 2) # OPCODE_BINARY

    def sendEOF(self):
        with self._lock:
            logging.debug("-- Sending EOF --")
            self._ws.send(b'meta:EOF', 2) #OPCODE_BINARY

class Connection(TunnelSocket):
    def __init__(self, info, listener=None):
        user,dev = parseDeviceId(info.devId)
        auth = config.getClientAuth(user)

        if auth[0] == user:
            # device belongs to local user (or at least to a user we have credentials for) -> use the unqualified devId
            devId = dev
        else:
            devId = info.devId
            
        #def __init__(self, endpoint, info, auth, listener=None, **params):
        TunnelSocket.__init__(self, '/connect', info, auth=auth, listener=listener, dev=devId, protocol=info.protocol, service=info.service)

    def _ping(self):
        logging.debug('sending tunnel ping')
        self._ws.send(b'meta:ping', 2) #OPCODE_BINARY
        TunnelSocket._ping(self)

class Response(TunnelSocket):
    def __init__(self, info, listener=None):
        auth = (config.getDeviceUser(), config.getDeviceAuth())
        TunnelSocket.__init__(self, '/accept', info, auth=auth, listener=listener, tunnel=info.tunnelId, key=info.devKey)

    def onEOF(self):
        """ Got an EOF from the remote host -> closing the websocket """
        with self._lock:
            self._ws.close()

    def _onClose(self, ws):
        state.add('tunnels', 'count', -1)
        state.remove('tunnels', 'info', self.info.connId)
        return TunnelSocket._onClose(self, ws)

    def _onConnected(self):
        state.add('tunnels', 'count', 1)
        state.set('tunnels.info', self.info.connId, self.info.__dict__)
        return TunnelSocket._onConnected(self)


def qualifyDeviceId(name):
    '''prepends the user name to unqualified device IDs
        - name: Device name (e.g. dev1, user0.dev1)

        returns: qualified deviceID (e.g. user0.dev1)
            if the name parameter was already qualified, it's returned as is.
    '''

    if '.' in name:
        return name
    elif '/' in name: # legacy user/device format (replace with the dot format)
        return '.'.join(name.split('/', 1))
    else:
        return '.'.join([config.getClientUser(), name])

def parseDeviceId(devId):
    '''Splits a (possibly) qualified device ID into its components

        - name:  Device name (e.g. dev1, user0.dev1)

        returns: (userName, deviceId)
          if the userName is not part of the name parameter, config.getClientUser() is returned
    '''
    user = None

    if '.' in devId: # format: user.device
        user, devId = devId.split('.', 1)
    elif '/' in devId: # legacy device IDs
        user, devId = devId.split('/', 1)
    else:
        user = config.getClientUser()

    return user, devId



"""
Queries the ondevice.io server for details on all your devices

Arguments:
--json
    Print each host's details as json objects, one host per line
--props (only in conjunction with --json)
    Also include device properties in the JSON output

Examples:
$ {cmd} list
ID          State               IP          Version     Title
foo.dev1    OFFLINE (for 1h)    1.2.3.4     0.1dev6     Device 1
foo.raspi   ONLINE (for 2d)     1.2.3.5     0.1dev6     Raspberry PI at home

$ {cmd} list --json --props
{{"stateTs": 1474530839829, "version": "ondevice v0.2.4", "state": "online", "id": "foo.raspi", "props": {{"hello": "world"}}, "name": "Raspberry PI"}}
{{"stateTs": 1474472379545, "version": "0.2.1", "state": "online", "id": "foo.otherDev", "props": {{}}, "name": "NAS"}}
"""

usage = {
    'msg': 'Displays detailed information on your devices',
    'group': 'client'
}

from ondevice.core import exception, rest, sock

import json
import getopt


def run(*args):
    # TODO add '--state=on/offline' option
    optlist, args = getopt.gnu_getopt(args, '', ['json', 'props'])
    optlist = dict(optlist)
    restUrl = '/devices'

    if len(args) > 0:
        raise exception.UsageError('Unsupported commandline arguments: {0}'.format(args))

    if '--props' in optlist:
        restUrl = '/devices?props=true'
    resp = rest.apiGET(restUrl)

    if '--json' in optlist:
        for dev in resp['devices']:
            print(json.dumps(dev))
#    elif '--raw' in args:
#        print(json.dumps(resp))
    else:
        fmt='{id:20s} {state:10s} {ip:15s} {version:15s} {name}'
        print(fmt.format(id='ID',state='State', ip='IP', version='Version', name='Title'))

        for dev in resp['devices']:
            if not 'name' in dev:
                dev['name'] = ''
            print(fmt.format(**dev))



from ondevice.core import config, sock, service, state
from ondevice import modules

import filelock
import json
import logging
import os
import psutil
import sys
import traceback
import time


class Daemon(sock.Socket):
	""" Runs the ondevice daemon """
	def __init__(self, sid=None):
		self._abortMsg = None
		self.sid = sid

		key = config.getDeviceKey()
		params = {}
		if key != None:
			params['id'] = key
		if sid != None:
			params['sid'] = sid

		auth = (config.getDeviceUser(), config.getDeviceAuth())
		sock.Socket.__init__(self, '/serve', auth=auth, **params)

	def onMessage(self, msg):
		try:
			if not '_type' in msg:
				raise Exception("Missing message type: {0}".format(msg))
			elif msg._type == 'hello':
				assert not self._connectionSucceeded
				logging.info("Connection established, online as '%s'", msg.name)
				state.setAll('device', state='online', devId=msg.name)
				config.setDeviceId(msg.name)
				self.key = msg.devId
				self.sid = msg.sid
				config.setDeviceKey(msg.devId)

				for name, svc in service.listServices().items():
					self.send({'_type': 'announce', 'name': name, 'protocol': svc['protocol']})

				self._connectionSucceeded = True
			elif msg._type == 'ping':
				# send back a 'pong' message
				logging.debug("Got ping: %s", repr(msg))
				response = {'_type': 'pong', 'ts': msg.ts}
				self.send(response)
			elif msg._type == 'connect':
				logging.info("Got '%s' request by user %s (ip: %s)", msg.service, msg.clientUser, msg.clientIp)
				try:
					errTpl = {'_type': 'connectError', 'tunnelId': msg.tunnelId}
					if not service.exists(msg.service):
						return self.sendConnError(404, "Service '{0}' not found".format(msg.service), msg.tunnelId)
					proto = service.get(msg.service)['protocol']
					if msg.protocol != proto:
						return self.sendConnError(400, "Protocol mismatch (expected={0}, actual={1})".format(proto, msg.protocol), msg.tunnelId)
					elif not modules.exists(proto):
						return self.sendConnError(404, "Module '{0}' not found!".format(msg.protocol), msg.tunnelId)
					svc = modules.getService(msg, self.key)

					svc.startRemote()
					svc.startLocal()
				except Exception as e:
					# got error -> notify the server
					self.send({'_type': 'connectError', 'tunnelId': msg.tunnelId, 'code': 502, 'msg': repr(e)})
					raise e

			elif msg._type == 'error':
				errType = 'Error'
				if msg.code == 400: errType = 'Bad Request'
				elif msg.code == 403: errType = 'Forbidden'
				elif msg.code == 404: errType = 'Not Found'

				logging.error("%s: %s", errType, msg.msg)
			else:
				logging.error("onMessage: unsupported type")
				logging.error("  msg=%s", msg)
		except Exception as e:
			# the websocket-client lib swallows the stack traces of exceptions
			print("Msg: '{0}'".format(msg))
			traceback.print_exc()
			raise e

	def _onError(self, ws, error):
		if type(error) is KeyboardInterrupt:
			self._abortMsg = "Keyboard interrupt, exiting"
			return

		sock.Socket._onError(self, ws, error)

	def run(self):
		self._connectionSucceeded = False
		sock.Socket.run(self)
		state.remove('device', 'state', 'devId')
		return self._connectionSucceeded

	def send(self, msg):
		data = json.dumps(msg)
		logging.debug('>> %s', data)
		self._ws.send(data)

	def send2(self, **kwargs):
		self.send(kwargs)

	def sendConnError(self, code, msg, tunnelId):
		logging.error("Connection error: {0} (code {1})".format(msg, code))
		self.send2(_type='connectError', tunnelId=tunnelId, code=code, msg=msg)

def getDaemonPID():
    pidfile = config._getConfigPath('ondevice.pid')
    if os.path.exists(pidfile):
        with open(pidfile, 'r') as f:
            rc = int(f.read().strip())
            if rc > 0 and psutil.pid_exists(rc):
				# TODO check if it's actually the right PID
                return rc

    return None

def runForever():
	""" Run the device endpoint in a loop (until it gets aborted) """
	lock = filelock.FileLock(config._getConfigPath('ondevice.lock'))
	try:
		with lock.acquire(timeout=0):
			_runForever()
	except filelock.Timeout:
		logging.error("There's already another instance of `ondevice daemon` running!")
		sys.exit(1)

def _runForever():
	""" runForever() internals without the lock file mechanism """
	retryDelay = 10
	sid = None

	pidfile = config._getConfigPath('ondevice.pid')
	with open(pidfile, 'w') as f:
		f.write('{0}\n'.format(os.getpid()))

	try:
		while (True):
			# TODO right now it's impossible to reuse Daemon objects (since the URL's set in the constructor but the device key might change afterwards)
			daemon = Daemon(sid=sid)
			if daemon.run() == True:
				state.setAll('device', state='reconnecting')
				retryDelay = 10
			if daemon._abortMsg != None:
				logging.info(daemon._abortMsg)
				break

			else:
				logging.info("Lost connection, retrying in %ds", retryDelay)
				time.sleep(retryDelay)
				retryDelay = min(900, retryDelay*1.5)
				sid = daemon.sid
	finally:
		os.unlink(pidfile)



from ondevice.core import state

import logging
import threading
import time

class ThreadInfo:
    """ Stores some information on each BackgroundThread instance """
    def __init__(self, **kwargs):
        for key in 'id,name'.split(','):
            if key not in kwargs:
                raise KeyError("Missing required ThreadInfo attribute: '{0}' (got: {1})", key, kwargs.keys())

        for k,v in kwargs.items():
            setattr(self, k, v)


class BackgroundThread():
    """ Thin wrapper around threading.Thread with simple event support.

    During the lifecycle of a backgrond thread the following events may occur (roughly in that order):
    - started: after a call to start()
    - running: before target() is being executed
    - stopping: after a call to stop()
    - finished: after target() has finished
    """
    def __init__(self, target, stopFn, name, args=(), kwargs={}):
        """ Create a BackgroundThread object
        - target: Thread function
        - stopFn: Function to gracefully stop the thread function (`target`) - will be called when invoking .stop()
        - name: Thread name (doesn't have to be unique)
        - args:  arguments to pass to target()
        - kwargs: keyword arguments to pass to target() """

        threadId = state.add('threads', 'seq', 1)
        self.info = ThreadInfo(name=name, id=threadId)

        self.target = target
        self.stopFn = stopFn
        self._listeners = {}
        self._thread = threading.Thread(target=self.run, name=name, args=args, kwargs=kwargs)

    def _emit(self, event, *args, **kwargs):
        listeners = list(self._listeners[event]) if event in self._listeners else []
        logging.debug("thread {0} fired event: {1} (args={2}, kwargs={3}, {4} listeners)".format(self.target, event, args, kwargs, len(listeners)))

        for l in listeners:
            l(*args, **kwargs)

    def run(self):
        """ Runs the target function (don't call this directly unless you want the target function be run in the current thread) """

        #
        state.add('threads', 'count', 1)
        self.info.startedAt = time.time()
        state.set('threads.info', self.info.id, self.info.__dict__)
        self._emit('running')
        try:
            self.target()
        finally:
            self._emit('finished')
            state.remove('threads.info', self.info.id)
            state.add('threads', 'count', -1)

    def addListener(self, event, fn):
        if event not in ['started', 'running', 'stopping', 'finished']:
            raise KeyError("Unsupported event: '{0}'".format(event))
        if not event in self._listeners:
            self._listeners[event] = set()

        self._listeners[event].add(fn)

    def addListenerObject(self, obj):
        """ Helper function to bind all signals to predefined methods of `obj` (if they exist)
         - mainly used for testing """
        fns = {'threadStarted': 'started', 'threadRunning':'running', 'threadStopping':'stopping', 'threadFinished':'finished'}
        for fn,event in fns.items():
            if hasattr(obj, fn):
                self.addListener(event, getattr(obj, fn))

    def removeListener(self, event, fn):
        if event not in self._listeners:
            return # Cant' remove something that isn't there
        self._listeners[event].remove(fn)

    def start(self):
        self._thread.start()
        self._emit('started')

    def stop(self):
        """ call the stopFn passed to the constructor and emit the 'stopping' signal """
        self.stopFn()
        self._emit('stopping')


class FixedDelayTask(BackgroundThread):
    """ Represents a repeating task that runs with a fixed delay (delay)
    in a background thread """

    def __init__(self, target, interval, name, *args, **kwargs):
        self._target = target
        self.interval = interval
        self.args = args
        self.kwargs = kwargs
        self._event = threading.Event()
        BackgroundThread.__init__(self, self._run, self._stop, name)

    def _run(self):
        while not self._event.wait(self.interval):
            try:
                self._target(*self.args, **self.kwargs)
            except:
                logging.exception('Exception in FixedDelayTask')

    def _stop(self):
        self._event.set()



"""
{cmd} service
  - lists local services
{cmd} service add [-h/--hidden] <protocol> <name> [parameters]
  - adds a (hidden) service with the given protocol, name and parameters
{cmd} service rm <name>
  - removes a service (by name)

The default installation only has the 'ssh' 'echo' service enabled.

Parameters:
- `--hidden`

Examples:

$ {cmd} service
Proto   Name    Params
ssh     ssh     --port=378
echo    [echo]

$ {cmd} service rm echo
$ {cmd} services add echo otherEcho --hidden
"""

usage = {
    'args': '[add/rm] [args...]',
    'msg': 'Manages services',
    'group': 'device',
    'hidden': True # hide it for now - since only 'ssh' is officially supported right now
}

from ondevice.core import exception, service


def addService(name, protocol):
    # TODO parse additional arguments ('--hidden', ...)
    service.add(name, protocol)


def listServices():
    for name, info in service.listServices().items():
        details = ['protocol={0}'.format(info['protocol'])]
        if 'hidden' in info and info['hidden'] == True:
            details.append('HIDDEN')
        print('{0}\t{1}'.format(name, ', '.join(details)))

def rmService(name):
    service.remove(name)

def run(subcmd='list', *args):
    if subcmd == 'list':
        listServices(*args)
    elif subcmd == 'add':
        addService(*args)
    elif subcmd == 'rm':
        rmService(*args)
    else:
        raise exception.UsageError("Unsupported subcommand: {0}".format(subcmd))



"""
Run the ondevice daemon.

Options:

-f
--foreground
    Don't run as background daemon but keep attached to the terminal

"""

usage = {
    'msg': 'Run the ondevice device daemon',
    'group': 'device'
}

from ondevice.core import config, daemon, exception
from ondevice import control

from daemon import DaemonContext

import getopt

def runDaemon():
    control.server.start()
    daemon.runForever()
    control.server.start()

def run(*args):
    opts, args = getopt.gnu_getopt(args, 'C:f', ('configDir=', 'foreground'))
    foreground = False

    if len(args) > 0:
        raise exception.UsageError("Extra arguments: {0}".format(args))

    for k,v in opts:
        if k in ['-C', '--configDir']:
            config.configDir = v
            config.invalidateCache()

        elif k in ['-f', '--foreground']:
            foreground = True

    if foreground:
        runDaemon()
    else:
        with DaemonContext():
            runDaemon()



"""
Set up the API keys for communicating with the ondevice.io service.

Example:
$ ondevice setup
User: ondevice
API key: **********
INFO:root:Updated client key (user: 'ondevice')
"""

usage = {
    'msg': 'Set up the API keys for communicating with the ondevice.io service'
}

from ondevice.core import config, rest, sock

import getpass
import logging
import six


def run():
    setupKeys()
    # TODO set up autostart

def setupKeys():
    # TODO check if the config file already has API keys (and ask if the user really wants to overwrite them)
    user,key = queryCredentials()

    config.overrides.put('client', 'user', user)
    config.overrides.put('client', 'auth', key)

    resp = rest.apiGET('/keyInfo')
    config.overrides.clear() # just to be sure, reset to default authentication

    roles = resp['roles']

    if 'client' in roles:
        config.setClientUser(user)
        config.setClientAuth(key)
        logging.info("Updated client key (user: '{0}')".format(user))
    if 'device' in roles:
        config.setDeviceUser(user)
        config.setDeviceAuth(key)
        logging.info("Updated device key (user: '{0}')".format(user))


def queryCredentials():
    user = six.moves.input('User: ')
    key = getpass.getpass('API key: ')
    return user,key



from ondevice.core import exception

import pkgutil

def listCommands():
    for importer, modname, ispkg in pkgutil.iter_modules(__path__):
        if not ispkg and modname.endswith('_cmd'):
            yield modname[:-4]

def load(cmdName):
    try:
        ondevice = __import__('ondevice.commands.{0}_cmd'.format(cmdName))
    except ImportError:
        raise exception.UsageError("Unknown command, try 'ondevice help' for a list of commands: {0}".format(cmdName))
    return getattr(ondevice.commands, '{0}_cmd'.format(cmdName))

def run(cmdName, *args, **opts):
    cmd = load(cmdName)
    return cmd.run(*args, **opts)

def usage(cmdName):
    cmd = load(cmdName)
    if not hasattr(cmd, 'usage'):
        raise exception.ImplementationError("Command '{0}' doesn't have 'usage' information".format(cmdName))
    if callable(cmd.usage):
        args, msg = cmd.usage()
        return {'args': args, 'msg': msg}
    else: return cmd.usage



try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser

from ondevice.core import exception

import os


# default config directory:
#   ~/.config/ondevice/
# This works well on macOS and Linux
# TODO add support for further OSs
configDir = os.path.join(os.path.expanduser('~'), '.config/ondevice')

class Overrides:
    def __init__(self):
        self.data = {}

    def __contains__(self, key):
        return key in self.data

    def clear(self):
        self.data = {}

    def get(self, section, key, default=None):
        if not (section,key) in self.data:
            return default
        return self.data[(section,key)]

    def items(self):
        for key, value in self.data.items():
            yield key[0], key[1], value

    def put(self, section, key, value):
        self.data[(section,key)] = value

    def remove(self, section, key):
        if (section, key) in self.data:
            del(self.data[(section,key)])

_config = None
overrides = Overrides()

def addSection(name):
    return _getConfig(True).add_section(name)

def getClientAuth(tgtUser=None):
    """ Returns a (username, password) client access token tuple.
    If you specify a tgtUser and your ondevice.conf contains matching credentials, those will be returned instead.
    """

    clientUser = getClientUser()
    if tgtUser != None:
        if tgtUser != clientUser:
            userKey = 'auth_{0}'.format(tgtUser)
            if hasValue('client', userKey):
                return tgtUser, getValue('client', userKey)
    clientKey = getValue('client', 'auth')
    if clientUser != None and clientKey != None:
        return clientUser, clientKey
    else: return None

def getClientUser():
    return getValue('client', 'user')

def getDeviceAuth():
    envAuth = os.getenv("ONDEVICE_AUTH")
    if envAuth != None:
        # allow the credentials to be provided using environment variables
        return envAuth
    else:
        return getValue('device', 'auth')


def getDeviceKey():
    return getValue('device', 'key')

def getDeviceId():
    # remove old device 'name' config value if present
    if hasValue('device', 'name'):
        remove('device', 'name')
    return getValue('device', 'dev-id')

def getDeviceUser():
    envUser = os.getenv("ONDEVICE_USER")
    if envUser != None:
        # allow the credentials to be provided using environment variables
        return envUser
    else:
        return getValue('device', 'user')

def hasSection(name):
    return _getConfig().has_section(name)

def hasValue(section, name):
    return _getConfig().has_option(section, name)

def listKeys(section):
    return _getConfig().options(section)

def setClientAuth(auth): setValue('client', 'auth', auth)
def setClientUser(name): setValue('client', 'user', name)
def setDeviceAuth(auth): setValue('device', 'auth', auth)
def setDeviceKey(key): setValue('device', 'key', key)
def setDeviceId(slug): setValue('device', 'dev-id', slug)
def setDeviceUser(name): setValue('device', 'user', name)

def invalidateCache():
    """ Invalidate the in-memory cache of the configuration
    (forcing a re-read the next time any getter or setter is called) """
    global _config
    _config = None

def _getConfig(reread=False):
    global _config
    if reread == True or _config == None:
        configFile = _getConfigPath('ondevice.conf')
        _config = ConfigParser()
        _config.read(configFile)

        # init sections
        for s in ['client', 'device']:
            if not _config.has_section(s):
                _config.add_section(s)
        if not _config.has_section('services'):
            _config.add_section('services')
            _config.set('services', 'ssh', '{"protocol": "ssh"}')

    return _config

def _getConfigPath(filename):
    global configDir

    if not os.path.isdir(configDir):
        try:
            os.makedirs(configDir)
        except OSError:
            raise exception.ConfigurationError("Couldn't create config directory: {0}".format(configDir)) 

    return os.path.join(configDir, filename)

def getValue(section, key, default=None):
    if (section,key) in overrides:
        return overrides.get(section, key, default)

    cfg = _getConfig()
    if cfg.has_option(section, key):
        return cfg.get(section, key)
    else:
        return default

def remove(section, key):
    cfg = _getConfig(True)
    if not cfg.remove_option(section, key):
        raise KeyError("No such value: {0}/{1}".format(section, key))

    _saveConfig()

def setValue(section, key, value):
    cfg = _getConfig(True)
    overrides.remove(section, key)
    cfg.set(section, key, value)
    _saveConfig()

def _saveConfig():
    cfg = _getConfig()
    cfgPath = _getConfigPath('ondevice.conf')
    tmpPath = os.path.join(os.path.dirname(cfgPath), '.{0}.tmp'.format(os.path.basename(cfgPath)))
    with open(tmpPath, 'w') as f:
        cfg.write(f)
    os.rename(tmpPath, cfgPath)



"""
Copy files from/to devices using rsync

Usage:

ondevice rsync [rsync-options...]

Examples:

- ondevice rsync -av /source/path/ root@myDev:/target/path/
    copy the local /src/path to myDev's /target/path/ as root
    (and pass the -a and -v options to rsync)
- ondevice rsync me@otherDev:/etc/motd /tmp/other.motd
    copy otherDev's /etc/motd file to /tmp/other.motd (and login as 'me')

This command is only a thin wrapper around the `rsync` client (using its `-e`
argument to make it use `ondevice ssh` internally).

"""

import subprocess
import sys

usage = {
    'msg': 'Copy files from/to your devices using rsync',
    'args': '[ssh-arguments...]',
    'group': 'client'
}

def run(*args):
    rsync = subprocess.Popen(['rsync', '-e', 'ondevice ssh'] + list(args), stdin=None, stdout=None, stderr=None)
    return rsync.wait()



"""
unix/tcp server socket used for communication with the ondevice daemon
"""

import ondevice
from ondevice.core import config
from ondevice.core import state as state_

import cherrypy
import json
import logging
import os
import threading
from six.moves import urllib

class ControlSocket:
    def _toJson(self, data):
        cherrypy.response.headers['Content-Type'] = 'application/json'
        return json.dumps(data).encode('utf8')

    @cherrypy.expose
    def index(self):
        return ""

    @cherrypy.expose
    def state(self, full=False):
        s = state_.getCopy()
        rc = {
            'version': ondevice.getVersion()
        }

        stateKeys = ['device']
        if full:
            stateKeys.extend(['tunnels'])

            if 'threads' in full:
                # print system threads instead of BackgroundThread instances
                rc['threads'] = []
                for t in threading.enumerate():
                    rc['threads'].append({'name': t.name, 'daemon': t.daemon, 'id': t.ident})
            else:
                stateKeys.append('threads')

        for key in stateKeys:
            if key in s:
                rc[key] = s[key]

        return self._toJson(rc)

_cherry = None

def _getSocketUrl():
    rc = None
    if 'ONDEVICE_HOST' in os.environ:
        rc = os.environ['ONDEVICE_HOST']
    if rc == None:
        rc = 'unix://'+config._getConfigPath('ondevice.sock')
    return rc


def start():
    global _cherry

    if _cherry != None:
        logging.info("Cherrypy has already been started")
        return

    socketUrl = _getSocketUrl()
    url = urllib.parse.urlparse(socketUrl)

    if url.scheme == 'unix':
        bindAddr = url.path
    elif url.scheme == 'http':
        bindAddr = (url.host, url.port)

    cherrypy.server.bind_addr = bindAddr
    cherrypy.tree.mount(ControlSocket(), '/')
    cherrypy.server.start()

    _cherry = cherrypy.server

def stop():
    cherrypy.server.stop()

    # remove unix socket file
    sock = urllib.parse.urlsplit(_getSocketUrl())
    if sock.scheme == 'unix':
        if os.path.exists(sock.path):
            os.unlink(sock.path)
    
    




from ondevice.core import config, exception

from six.moves import urllib
import requests_unixsocket
import requests

ONDEVICE_HOST='http+unix://'+urllib.parse.quote(config._getConfigPath("ondevice.sock"), safe='')

_session = None

class Data:
    """ Slim wrapper around dict to make its items accessible via getattr() """
    def __init__(self, data):
        assert type(data) == dict
        self._data = data

    def __getattr__(self, key):
        rc = self._data[key]
        if type(rc) == dict:
            rc = Data(rc)
        return rc

    def __contains__(self, key):
        return key in self._data and self._data[key] != None

    def __repr__(self):
        return "Data({0})".format(repr(self._data))


def _getSession():
    """ Open and return a session on first use """
    global _session
    if _session == None:
        _session = requests_unixsocket.Session()
    return _session

def _makeUrl(endpoint, **params):
    global ONDEVICE_HOST
    parts = list(urllib.parse.urlsplit(ONDEVICE_HOST))
    parts[2] = endpoint

    if len(params) > 0:
        parts[3] = urllib.parse.urlencode(params)

    return urllib.parse.urlunsplit(parts)

def _getJson(endpoint, **params):
    try:
        url = _makeUrl(endpoint, **params)
        r = _getSession().get(url)
        if r.status_code >= 300:
            raise exception.TransportError("ondevice daemon responded with code {0}".format(r.status_code))
        return Data(r.json())
    except requests.exceptions.ConnectionError as e:
        raise exception.TransportError("failed to query ondevice daemon: {0}".format(e.args))


def getState(full=False):
    if full==False:
        return _getJson('/state')
    else:
        return _getJson('/state', full=full)



"""
Attempts to initiate a connection to a device's service.

There is actually a shorthand. Instead of `ondevice connect <module> ...` you
can write:

    `ondevice :<module> ...`

Arguments:
- module:   The module to use for the connection (e.g. 'ssh')
- dev:      Device to connect to
- svcName:  Service name (defaults to the module name)

Examples:
    {cmd} connect ssh someDevice - connect to the 'ssh' service of 'someDevice'
    {cmd} :ssh someDevice - same as the above (i.e. omitted 'connect')
    {cmd} :ssh@otherSsh someDevice - connect to the 'otherSsh' service using ssh
"""

usage = {
    'args': '<module> <dev> [svcName]',
    'msg': 'Connects to a service on the specified device (shorthand: `:<module>`)',
    'group': 'client',
    'hidden': True # hide this for now - since there's a proper `ondevice ssh` command now
}

from ondevice import modules


def run(module, dev, *args):
    client = modules.loadClient(devId=dev, protocolStr=module, args=args)
    client.startRemote()

    args = ()
    if hasattr(client, '_args'): args = client._args
    rc = client.runLocal(*args)
    return rc



"""Manage device properties

- {cmd} device <devName> props <devName> [key]... - List a device's properties
- {cmd} device <devName> set <devName> <key>=<value>... - set one or more device properties
- {cmd} device <devName> rm <key>... - remove specific device properties

Examples:
$ {cmd} device someDevice props
foo=bar
abc=123

$ {cmd} device someDevice props foo hello
foo=bar
hello=

$ {cmd} device someDevice rm abc
foo=bar

$ {cmd} device someDevice set hello=world answer=42
foo=bar
hello=world
answer=42
"""

usage = {
    'args': '<devName> props/set/rm [args]',
    'msg': "Fetch/Manage a device's information",
    'group': 'client'
}


from ondevice.core import exception, rest, sock


def _printProps(resp):
    for devId, devProps in resp['props'].items():
        for key,value in devProps.items():
            print("{0}={1}".format(key,value))


def propList(devName, *keys):
    resp = rest.apiGET("/device/{0}/props".format(devName))
    _printProps(resp)

def propSet(devName, *values):
    props = {}
    for val in values:
        k,v = val.split('=')
        props[k] = v

    resp = rest.apiPOST("/device/{0}/props".format(devName), data={'props': props})
    _printProps(resp)

def propRm(devName, *keys):
    resp = rest.apiDELETE("/device/{0}/props".format(devName), data={'props': keys})
    _printProps(resp)

def run(devName, subcmd, *args):
    if subcmd == 'props':
        propList(devName, *args)
    elif subcmd == 'set':
        propSet(devName, *args)
    elif subcmd == 'rm':
        propRm(devName, *args)
    else:
        raise exception.UsageError("Unsupported subcommand: {0}".format(subcmd))



""" Some exception classes for the ondevice client """

class _Exception(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        self.msg = args[0]
        for k,v in kwargs.items():
            setattr(self, k, v)

class ConfigurationError(_Exception):
    """ Indicates a missing/faulty configuration value """
    pass

class ImplementationError(_Exception):
    """ Indicates implementation issues with a command or module """
    pass

class TransportError(_Exception):
    """ Indicates a communication error with the server """
    pass

class UsageError(_Exception):
    """ Indicates issues with the commandline usage (unknown command, unsupported argument, etc.) """
    pass



""" stores non-persistent application state.

All functions in this module are thread safe """

import copy
import threading

_state = {}
_lock = threading.Lock()

def _getPath(path, create=False):
    global _state
    rc = _getPathRec(_state, path.split('.'), create)
    if rc == None:
        raise KeyError("state path not found: {0}".format(path))
    return rc

def _getPathOrNull(path):
    global _state
    return _getPathRec(_state, path.split('.'), False)

def _getPathRec(state, path, create=False):
    if len(path) == 0:
        return state
    key = path[0]
    if not key in state:
        if create:
            state[key] = {}
        else:
            return None

    return _getPathRec(state[key], path[1:], create)

def add(path, key, value):
    """ increment a state state value by a given integer (can be negative) """
    global _lock
    with _lock:
        parent = _getPath(path, True)
        if key not in parent:
            parent[key] = 0 # initialize with 0
        parent[key] += value
        return parent[key]

def getCopy():
    """ Returns a deep copy of the current state """
    global _lock, _state
    with _lock:
        return copy.deepcopy(_state)

def remove(path, *keys):
    """ Removes state keys, returning the number of keys actually removed """
    rc = 0

    global _lock
    with _lock:
        parent = _getPathOrNull(path)
        if parent != None:
            for key in keys:
                if key in parent:
                    del parent[key]
                    rc += 1

    return rc

def set(path, key, value):
    global _lock
    with _lock:
        parent = _getPath(path, True)
        parent[key] = value

def setAll(path, **kwargs):
    global _lock
    with _lock:
        parent = _getPath(path, True)
        for k,v in kwargs.items():
            parent[k] = v



#!/usr/bin/env python3

import getopt
import logging
import pkg_resources
import os, sys

from ondevice import commands
from ondevice.core import exception as ex

def _main():
    try:
        return main(sys.argv[1:])
    except ex.UsageError as e:
        logging.error("Usage error: {0}".format(e.args[0]))
        return 1
    except getopt.GetoptError as e:
        logging.error("Usage error: {0}".format(e.args[0]))
        return 1
    except KeyboardInterrupt:
        logging.error(" -- keyboard interrupt --")
        return 1

def main(args):
    logging.basicConfig(level=logging.INFO)

    if len(args) < 1:
        usage()
    else:
        opts = parseArgs(args)

        cmd = args.pop(0)
        if cmd.startswith(':'):
            args.insert(0, cmd[1:])
            cmd = 'connect'

        return commands.run(cmd, *args, **opts)

def getVersion():
    pkgInfo = pkg_resources.require("ondevice")[0]
    return pkgInfo.version


def parseArgs(args, aliases={}):
    """ takes all options (anything starting with a -- or a -) from `args`
    up until the first non-option.
    short options (those with only one dash) will be converted to their long
    version if there's a matching key in `shortOpts`. If there isn't, an
    KeyError is raised."""
    rc = {}

    while len(args) > 0 and args[0].startswith('-'):
        opt = args.pop(0)
        if opt == '--':
            break # ignore and exit

        # treat options with one dash the same as those with two dashes
        if opt.startswith('--'):
            opt = opt[2:]
        elif opt.startswith('-'):
            opt = opt[1:]

        equalsPos = opt.find('=')
        val = None
        if equalsPos >= 0:
            val = opt[equalsPos+1:]
            opt = opt[:equalsPos]

        rc[opt] = val

    return rc

def usage(exitCode=1):
    commands.run('help')
    sys.exit(exitCode)

if __name__ == '__main__':
    sys.exit(_main())



"""
Connect to your devices using the 'ssh' command.

Usage:
    ondevice ssh [<user>@]<device> [ssh-arguments...]

This is a relatively thin wrapper around the `ssh` command.
The main difference to invoking ssh directly is that instead of regular host names you'll have to specify an ondevice deviceId.
The connection is routed through the ondevice.io network.

ondevice ssh will try to parse ssh's arguments, the first non-argument has to be
the user@hostname combo.

See ssh's documentation for further details.

Examples:
- ondevice ssh device1
    simply connect to device1
- ondevice ssh user@device1
    open an SSH connection to device1, logging in as 'user'
- ondevice ssh device1 echo hello world
    run 'echo hello world' on device1
- ondevice ssh device1 -N -L 1234:localhost:80
    Tunnel the HTTP server on device1 to the local port 1234 without opening
    a shell
- ondevice ssh device1 -D 1080
    Starting a SOCKS5 proxy listening on port 1080. It'll redirect all traffic
    to the target host.

"""

from ondevice.core import tunnel

import getopt
import subprocess
import sys

usage = {
    'msg': 'Connect to your devices using the ssh protocol',
    'args': '[ssh-arguments...]',
    'group': 'client'
}

def run(*args):
    # option list copied from debian jessie's openssh source package (from ssh.c, line 509)
    opts, nonopts = getopt.getopt(args, '1246ab:c:e:fgi:kl:m:no:p:qstvx')
    nonopts = list(nonopts)

    devId = nonopts.pop(0)
    sshUser = None
    if '@' in devId:
        sshUser, devId = devId.split('@', 1)

    # prepare command
    cmd = ['ssh', '-oProxyCommand={0} connect ssh:tunnel {1}'.format(sys.argv[0], devId)]

    # add options
    for k,v in opts:
        cmd.append('{0}{1}'.format(k,v))

    # add hostname (and user) in the format user@ondevice:devUser.devId
    qualifiedId = tunnel.qualifyDeviceId(devId)
    hostname = 'ondevice:{0}'.format(qualifiedId)
    if sshUser != None:
        hostname = '{0}@{1}'.format(sshUser, hostname)
    cmd.append(hostname)

    ssh = subprocess.Popen(cmd + nonopts, stdin=None, stdout=None, stderr=None)
    return ssh.wait()



"""
Attempts to gracefully stop a running ondevice daemon

After sending the terminate signal, it'll wait for up to five seconds for the
daemon to stop.
Only if the daemon was running, it'll return 0

Returns:
- 0 on success
- 1 on exceptions (if you encounter those, let us know)
- 2 if the daemon wasn't running
- 3 on timeout
- >3 might be used in the future
"""

usage = {
    'msg': 'Stop a running ondevice daemon',
    'group': 'device'
}

from ondevice.core import daemon

import os
import psutil
import signal
import sys
import time

def run():
    pid = daemon.getDaemonPID()
    if pid != None:
        sys.stderr.write("Stopping ondevice daemon...")
        sys.stderr.flush()

        # effective timeout: 5sec (20*250ms)
        for i in range(20):
            if i % 4 == 0:
                # send a SIGTERM once a second
                try:
                    os.kill(pid, signal.SIGTERM)
                except ProcessLookupError as e:
                    sys.stderr.write("\nondevice daemon isn't running (stale pid: {0})\n".format(pid))
                    return 2

            if not psutil.pid_exists(pid):
                sys.stderr.write("done\n")
                return 0 # stopped it successfully

            sys.stderr.write('.')
            sys.stderr.flush()
            time.sleep(0.25)

        sys.stderr.write("\nERROR: Timeout while trying to stop ondevice daemon (pid: {0})\n".format(pid))
        return 3

    else:
        sys.stderr.write("ERROR: ondevice daemon isn't running\n")
        return 2




# TODO implement proper service handling (persisting the active services and giving the user a way to register/remove them easily)

from ondevice.core import config, exception
#from ondevice import modules

import json
import re

def add(name, protocol, hidden=False, **options):
    data = options
    data['protocol'] = protocol
    if hidden == True:
        data['hidden'] = True

    if not re.match('[a-zA-Z0-9_]+', name):
        raise exception.UsageError("Illegal characters in service name: '{0}'".format(name))

    if config.hasValue('services', name):
        raise exception.UsageError("Service '{0}' already exists!".format(name))
    # TODO have some nicer error messages
    from ondevice import modules # TODO find a better way to import this (issue: circular imports)
    modules.load(protocol)

    config.setValue('services', name, json.dumps(data))

def exists(name):
    from ondevice import modules
    svcs = listServices()
    if name in svcs:
        return modules.exists(svcs[name]['protocol'])

def get(name):
    return listServices()[name]

def listServices():
    rc = {}
    for svcName in config.listKeys('services'):
        rc[svcName] = json.loads(config.getValue('services', svcName))

    return rc

def remove(name):
    config.remove('services', name)



""" Simple echo module (responds with the same data it gets)

The idea of this module is to provide both a test module to test connections and
an example module for you to base your own module implementations on.

See the [yet to be written - the module infrastructure is not considered stable
yet] module implementation guide.
"""

from ondevice.modules import ModuleClient, ModuleService

import sys

# short module description that will be shown when users type `ondevice modules`
info = 'Simple test module that sends back what it receives'
# Indicates whether or not the underlying protocol is itself encrypted.
encrypted = False

class Client(ModuleClient):
    def onClose(self):
        self.getConsoleBuffer(sys.stdin).close()

    def onEOF(self):
        self.getConsoleBuffer(sys.stdout).close()

    def onMessage(self, data):
        stream = self.getConsoleBuffer(sys.stdout)
        stream.write(b"> "+data)
        stream.flush()

    def runLocal(self):
        stream = self.getConsoleBuffer(sys.stdin)
        while True:
            data = stream.readline()
            if data:
                self._conn.send(data)
            else:
                self._conn.sendEOF()
                return

    def stopLocal(self):
        """ Closes stdin to get out of the runLocal() loop """
        self.getConsoleBuffer(sys.stdin).close()

class Service(ModuleService):
    def runLocal(self):
        pass

    def stopLocal(self):
        pass

    def onMessage(self, data):
        self._conn.send(data)



"""
{cmd} module [list]
    list the modules that are currently available.

{cmd} module info <name>

Modules have a device- and a client side implementation and essentially tunnel a
data in certain protocol through the ondevice.io network.
"""

# TODO enable this part of the documentation once user-installed modules are implemented
#There are a couple of modules already distributed with ondevice, but you can write
#your own and place them in ~/.config/ondevice/modules/ (of the user that runs
#ondevice)
#"""

usage = {
    'msg': 'Lists installed modules',
    'hidden': True # hide for now - until other, non-ssh based protocols are officially supported
}

from ondevice import modules
from ondevice.core import exception

import sys


def moduleInfo(name):
    mod = modules.load(name)
    print(mod.__doc__)


def listModules():
    for name in modules.listModules():
        mod = modules.load(name)
        info = '-- missing module information! --'
        if hasattr(mod, 'info'):
            info = mod.info
        print("{0}\t{1}".format(name, info))


def run(subcmd='list', *args):
    if subcmd == 'info':
        return moduleInfo(*args)
    elif subcmd == 'list':
        return listModules(*args)
    else:
        raise exception.UsageError("Unknown subcommand: {0}".format(subcmd))



"""
Lists available commands or prints detailed help for one of them

Examples:
    {cmd} help          lists available commands
    {cmd} help listen   shows the help for the 'listen' command
    {cmd} help help     shows this message
"""

usage = {
    'args': '[cmd]',
    'msg': 'lists available commands or prints detailed help for one'
}

from ondevice import commands
from ondevice.core import exception

import sys

def run(cmdName=None):
    if cmdName == None:
        print("USAGE: {0} <command> [args]\n".format('ondevice'))

        cmds = {'client':[], 'device':[], None:[]}
        groupNames = {'client': 'Client commands', 'device': 'Device commands', None: 'Other commands'}
        for cmd in commands.listCommands():
            usage = commands.usage(cmd)
            usage.setdefault('cmd', cmd)
            usage.setdefault('args', '')
            group = usage.setdefault('group', None)
            hidden = usage.setdefault('hidden', False)

            if group not in cmds.keys():
                group = None
            if not hidden:
                cmds[group].append(usage)

        for group in ['device', 'client', None]:
            print("- {0}:".format(groupNames[group]))
            for usage in cmds[group]:
                print("    {cmd} {args}\n\t{msg}".format(**usage))
            print('')
    else:
        cmd = commands.load(cmdName)

        if not hasattr(cmd, '__doc__') or cmd.__doc__ == None:
            raise exception.ImplementationError("Command '{0}' lacks documentation!".format(cmdName))

        usage = commands.usage(cmdName)
        usage.setdefault('cmd', cmdName)
        usage.setdefault('args', '')
        print('{0} {cmd} {args}'.format('ondevice', **usage))
        print(cmd.__doc__.format(cmd='ondevice'))



from ondevice.core import exception
import ondevice

from six.moves import urllib
import base64
import json
import logging
import ssl
import websocket

BASE_URL=urllib.parse.urlparse('wss://api.ondevice.io/v1.1')
#BASE_URL=urllib_parse.urlparse('ws://localhost:8080/v1.1')
#BASE_URL=urllib.parse.urlparse('wss://local.ondevice.io:8443/v1.1')
class Message:
    def __init__(self, data):
        assert type(data) == dict
        self._data = data

    def __contains__(self, key): return key in self._data
    def __getattr__(self, key): return self._data[key]
    def __repr__(self): return "Message({0})".format(self._data)

class Socket:
    """ Represents a websocket connection (to /serve, /connect or /accept) """
    def __init__(self, endpoint, auth, baseUrl=None, **params):
        """ Creates a Socket instance
        - endpoint: API endpoint (the part of the URL after baseUrl, e.g. '/serve')
        - auth: a username/password string tuple
        - baseUrl: replaces the BASE_URL if specified (necessary for '/accept' calls since those have to connect to the same API server as their '/connect' counterparts)
        - **params: URL params
        """
        global BASE_URL

        # TODO make base URL configurable
        if baseUrl == None:
            baseUrl = BASE_URL.geturl()

        paramStr = urllib.parse.urlencode(params)
        self._url = '{baseUrl}{endpoint}/websocket?{paramStr}'.format(**locals())

        headers = []
        headers.append('User-agent: ondevice v{0}'.format(ondevice.getVersion()))
        if auth != None:
            basicAuth = "{0}:{1}".format(*auth).encode('ascii')
            headers.append("Authorization: Basic {0}".format(base64.b64encode(basicAuth).decode('ascii')))
        else:
            raise exception.ConfigurationError("Missing authentication data")

        self._ws = websocket.WebSocketApp(self._url,
            header=headers,
            on_message=self._onMessage,
            on_error=self._onError,
            on_open=self._onOpen,
            on_close=self._onClose)

    def _getSslVersion(self):
        """ In python2.6 the protocol setting PROTOCOL_SSLv23 doesn't include TLS.
        This method works around that issue by only selecting PROTOCOL_SSLv23 if
        PROTOCOL_TLSv1_1 is available """
        if hasattr(ssl, 'PROTOCOL_TLSv1_1'):
            # Python >= 2.7.9 + openssl >= 1.0.1 -> SSLv23 is safe
            return ssl.PROTOCOL_SSLv23
        else:
            # there's no TLSv1.1+ support -> force TLSv1.0
            return ssl.PROTOCOL_TLSv1

    def _onMessage(self, ws, messageText):
        msg = Message(json.loads(messageText))
        logging.debug("<< %s", messageText)
        self.onMessage(msg)

    def _onClose(self, ws):
        logging.debug("onClose:")
        logging.debug("  ws=%s", ws)

    def _onError(self, ws, error):
        logging.error("onError:")
        logging.error("  ws=%s", ws)
        logging.error("  error=%s", error)

    def _onOpen(self, ws):
        logging.debug("onOpen:")
        logging.debug("  ws=%s", ws)
        if (hasattr(self, 'onOpen')):
            self.onOpen(self)

    def close(self):
        self._ws.close()

    def run(self):
        # Python2.6 fix (its ssl module won't try TLSv1 unless we explicitly tell it to)
        sslopt = {'ssl_version': self._getSslVersion()}

        self._ws.run_forever(sslopt=sslopt)

    def send(self, data):
        self._ws.send(data)



from unittest import TestCase
from ondevice.core import thread

import threading
import time

class GenericListener:
	""" Class that logs all method calls to self._log """
	class Method:
		def __init__(self, name, logFn):
			self.name = name
			self.logFn = logFn

		def __call__(self, *args, **kwargs):
			print("call to {0}:".format(self.name), args, kwargs)
			self.logFn({'name': self.name, 'ts': time.time(), 'args': args, 'kwargs': kwargs})

	def __init__(self):
		self._log = []

	def __getattr__(self, name):
		method = GenericListener.Method(name, self._log.append)
		setattr(self, name, method)
		return method

	def _hasEach(self, *args, **kwargs):
		""" returns True if each of the events from *args has occured once and each in **kwargs exactly the number of times of each param value """
		# convert args to kwargs ('foo' becomes {'foo': 1})
		for arg in args:
			if not arg in kwargs:
				kwargs[arg] = 1

		# go through the events we've capture
		data = {}
		for l in self._log:
			name = l['name']
			if not name in data: data[name] = 0
			data[name] += 1

		# compare them
		for event,count in kwargs.items():
			if event not in data:
				raise Exception("missing event: '{0}' ({1})".format(event, data))
			elif data[event] != count:
				raise Exception("event count mismatch (event={0}): expected={1}, actual={2}".format(event, count, data[event]))
			del data[event]

		# check unexpected events
		for ev, count in data.items():
			raise Exception("Got unexpected {0} events (count: {1})".format(ev, count))

class BackgroundThreadTest(TestCase):
	def _createThread(self, target, stopFn=None):
		""" creates and initializes a BackgroundThread instance with matching listener """
		listener = GenericListener()
		if target == None:
			target = listener.target
		if stopFn == None:
			stopFn = listener.stopFn
		rc = thread.BackgroundThread(target, stopFn, 'someThread')
		rc.addListenerObject(listener)
		rc._log = listener
		return rc

	def testSimpleTask(self):
		""" starts a BackgroundThread and makes sure its event handlers are called in the expected order """
		t = self._createThread(None)
		t.start()

		beforeStop = time.time()
		t.stop()

		t._log._hasEach('threadStarted', 'threadRunning', 'threadFinished', 'threadStopping', 'target', 'stopFn')

		return True


	def testStartStopTask(self):
		""" starts a task that'll run until it's being stopped manually """
		lock = threading.Lock()
		lock.acquire()
		# the second acquire call in the thread will block until we call release
		t = self._createThread(lock.acquire, lock.release)
		t.start()

		time.sleep(0.2)
		t._log._hasEach('threadStarted', 'threadRunning')

		t.stop()
		time.sleep(0.2)
		t._log._hasEach('threadStarted', 'threadRunning', 'threadStopping', 'threadFinished')

	def testStartFinishTask(self):
		""" same as above but the thread stops by itself (i.e. no 'stopping' event is triggered) """
		lock = threading.Lock()
		lock.acquire()

		t = self._createThread(lock.acquire)
		t.start()

		time.sleep(0.2)
		lock.release() # causes the thread target function to finish
		time.sleep(0.2)

		t._log._hasEach('threadStarted', 'threadRunning', 'threadFinished')

# TODO find a way to test FixedDelayTask in a predictable way



"""
Prints the client version and ondeviceID (if the daemon is running).

Note that this command will include further information (like the daemon version
in case we're querying a non-local daemon) in the future.

Returns:
- 0 if the daemon is running and online
- 1 if the daemon is running, but offline/reconnecting
- 2 if the daemon is not running
- >2 on errors
"""

import ondevice
from ondevice.core import daemon, exception
from ondevice.control import client

import getopt
import json
import logging

usage = {
    'msg': 'prints information on the ondevice client and daemon',
}

def run(*args):
    jsonOutput = False
    rc = 2 # assume it's not running

    # parse commandline arguments
    opts, args = getopt.gnu_getopt(args, '', ('json'))
    for k,v in opts:
        if k == '--json':
            jsonOutput = True

    state = {
        'client': {
            'version': ondevice.getVersion()
        }
    }
    
    try:
        s = client.getState(False)
        s._data.update(state)
        state = s
        
        rc = 1 # daemon is responding

        if 'device' in state:
            if state.device.state == 'online':
                rc = 0

    except exception.TransportError as e:
        # TODO actually check if the ondevice daemon is running (or if it was some other kind of TransportError)
        rc = 2
    except Exception as e:
        # TODO improve error handling
        logging.error("Caught exception")
        rc = 2

    # output
    if jsonOutput:
        print(json.dumps(state._data, indent=4))
    else:
        if 'device' in state:
            print("Device:")
            print("  ID: {0}".format(state.device.devId))
            print("")
        print("Client:")
        print("  version: {0}".format(ondevice.getVersion()))

    return rc